#include <iostream.h>
#include <strstream.h>
#include <fstream.h>
#include <time.h>
#include "ap.h"
#include "apint.h"

// Largest prime in primetable
const size_t MAXPRIME = 65537;

// Number of times to try the probable prime test before we are convinced
// Chances of the tested probable prime actually being composite is about
// one to 4^POWTESTS, that is for POWTESTS = 20 it is about 10e-12
const int POWTESTS = 20;

// Test for probable prime, the Miller-Rabin test
// For primes a^(p-1) == 1 (mod p) for all a
// For non-primes this is not true for most a
bool powtest (apint m)
{
    size_t t;

    for (t = 0; t < POWTESTS; t++)
        if (powmod (apint ((int) primetable[t]), m - 1, m) != 1)
            return false;

    return true;
}

// Filter out numbers which have n smallest primes as factors,
// using Erasthothenes' sieve
void sieve (apint m, bool *b, size_t n)
{
    size_t t, s, p;
    apint r;

    for (t = 0; t < n; t++)
        b[t] = true;

    for (t = 0; t < n; t++)
    {
        p = (size_t) primetable[t + 1];         // Start with 3, assume m is odd
        r = m % p;

        s = (size_t) ap2rawtype (r.val.ap);
        if (s) s = p - s;

        for (; s < n; s += p)
            b[s] = false;

        if (p == MAXPRIME) break;
    }
}

int main (int argc, char *argv[])
{
    size_t i, t, k, n;
    apint p, q, e, d, pq, pq1;
    char *c;
    bool *b, shorte = false;

    if (argc < 4)
    {
        cerr << "USAGE: rsagen keylength publickeyfile privatekeyfile [shortpublickey]" << endl;
        return 2;
    }

    if (!((istrstream (argv[1])) >> k))
    {
        cerr << "Invalid argument keylength: " << argv[1] << endl;
        return 1;
    }

    if (argc > 4)
        shorte = true;

    k /= 2;                     // k is the number of digits in one prime

    n = 50 * k;

    c = new char[2 * k];
    b = new bool[n];

    cout << "Input at least " << 2 * k << " random characters and press Enter: ";
    for (t = 0; t < 2 * k; t++)
    {
        i = (size_t) cin.get ();
        c[t] = (i + 69 * t + time (0)) % 10 + '0';      // Add pseudorandom noise
    }
    c[k - 1] |= 1;              // Make sure p will be odd
    c[2 * k - 1] |= 1;          // Make sure q will be odd

    istrstream o1 (c, k);
    istrstream o2 (c + k, k);

    o1 >> p;                    // Read initial p
    o2 >> q;                    // Read initial p

    cout << "Sieving candidates for p..." << endl;
    sieve (p, b, n);            // Filter candidates for prime p

    cout << "Testing candidates for p..." << endl;
    for (t = 0; t < n; t++, p += 2)     // Seek a most-likely-prime p from the candidates
        if (b[t] && powtest (p))
            break;

    if (t < n)
    {
        cout << "Suitable p was found." << endl;
    }
    else
    {
        cout << "Suitable p was not found. Please try again." << endl;
        return 1;
    }

    cout << "Sieving candidates for q..." << endl;
    sieve (q, b, n);            // Filter candidates for prime q

    cout << "Testing candidates for q..." << endl;
    for (t = 0; t < n; t++, q += 2)     // Seek a most-likely-prime q from the candidates
        if (b[t] && powtest (q))
            break;

    if (t < n)
    {
        cout << "Suitable q was found." << endl;
    }
    else
    {
        cout << "Suitable q was not found. Please try again." << endl;
        return 1;
    }

    pq = p * q;
    pq1 = (p - 1) * (q - 1);

    if (shorte)
    {
        istrstream i (argv[4]);

        // Generate a public key that is as small as possible to speed up encoding
        e = 3;
        i >> e;                     // Take it from the command line if specified
        if (e < 3) e = 3;
        else e += 1 - e % 2;        // Make sure e will be odd
    }
    else
    {
        // Add pseudorandom noise for generating e
        for (t = 0; t < 2 * k; t++)
            c[t] = (c[t] + 57 * t + time (0)) % 10 + '0';

        c[2 * k - 1] |= 1;          // Make sure e will be odd

        istrstream o3 (c, 2 * k);

        o3 >> e;                    // Read initial e

        e %= pq1;                   // e should be < p*q
    }

    cout << "Finding e..." << endl;
    for (t = 0; t < n; t++, e += 2)     // Seek an e which is relatively prime to pq1
    {
        apint a0 = pq1, a1 = e, q, r, t, y0 = 0, y1 = 1;

        q = a0 / a1;
        r = a0 - q * a1;

        while (r.sign ())               // Extended Euclidean Algorithm
        {
            t = y0 - q * y1;
            t %= pq1;
            if (t.sign () < 0) t += pq1;
            y0 = y1;
            y1 = t;
            a0 = a1;
            a1 = r;
            q = a0 / a1;
            r = a0 - q * a1;
        }

        if (a1 == 1)                    // gcd (e, pq1) == 1, multiplicative inverse exists
        {
            d = y1;                     // d is the multiplicative inverse of e mod pq1
            break;
        }
    }

    // Sanity check
    assert (gcd (e, pq1) == 1);
    assert (e * d % pq1 == 1);

    if (t < n)
    {
        cout << "Suitable e and d were found." << endl;
    }
    else
    {
        cout << "Suitable e was not found. Please try again." << endl;
        return 1;
    }

    cout << "Writing keys to output files..." << endl;
    ofstream pub (argv[2]);

    if (!pub)
    {
        cerr << "Unable to open public key file: " << argv[2] << endl;
        return 1;
    }

    ofstream priv (argv[3]);

    if (!priv)
    {
        cerr << "Unable to open private key file: " << argv[3] << endl;
        return 1;
    }

    pub << pq << endl << e << endl;
    priv << pq << endl << d << endl;

    cout << "Done." << endl;

    delete[] b;
    delete[] c;

    return 0;
}
